"""
    Tasks Executor
    ==============

    Responsible for executing tasks code, serialization, process separation, privileges management, passing the
    already prepared context, processing the results, processing declarative code

    Responsibilities are split into separate sub-layers.

"""
