Hijo del pueblo, te oprimen cadenas,
y esa injusticia no puede seguir;
si tu existencia es un mundo de penas
antes que esclavo prefiere morir.
En la batalla, la hiena fascista.
por nuestro esfuerzo sucumbirá;
y el pueblo entero, con los anarquistas,
hará que triunfe la libertad.

Trabajador, no más sufrir,
el opresor ha de sucumbir.
Levántate, pueblo leal,
al grito de revolución social.
Fuerte unidad de fe y de acción
producirá la revolución.
Nuestro pendón uno ha de ser:
sólo en la unión está el vencer.
