{% extends "./example-to-extend.j2" %}

{%- block WCH -%}
4 Jun 1862 textile worker and one of the founders of the Spanish anarchist movt, Teresa Claramunt was born. Jailed several times and played a leading role in the 1911 Aragon general strike. When she died 50k workers went to her funeral
{%- endblock -%}
